import re
from collections.abc import Callable
from pathlib import Path

import boto3
import click
from boto3 import Session

from dbt_platform_helper.constants import PLATFORM_CONFIG_FILE
from dbt_platform_helper.domain.maintenance_page import MaintenancePageProvider
from dbt_platform_helper.exceptions import ApplicationNotFoundError
from dbt_platform_helper.exceptions import AWSException
from dbt_platform_helper.utils.application import Application
from dbt_platform_helper.utils.application import load_application
from dbt_platform_helper.utils.aws import Vpc
from dbt_platform_helper.utils.aws import get_connection_string
from dbt_platform_helper.utils.aws import get_vpc_info_by_name
from dbt_platform_helper.utils.aws import wait_for_log_group_to_exist
from dbt_platform_helper.utils.files import apply_environment_defaults
from dbt_platform_helper.utils.messages import abort_with_error
from dbt_platform_helper.utils.validation import load_and_validate_platform_config


class DatabaseCopy:
    def __init__(
        self,
        app: str,
        database: str,
        auto_approve: bool = False,
        load_application: Callable[[str], Application] = load_application,
        vpc_config: Callable[[Session, str, str, str], Vpc] = get_vpc_info_by_name,
        db_connection_string: Callable[
            [Session, str, str, str, Callable], str
        ] = get_connection_string,
        maintenance_page_provider: Callable[
            [str, str, list[str], str, str], None
        ] = MaintenancePageProvider(),
        input: Callable[[str], str] = click.prompt,
        echo: Callable[[str], str] = click.secho,
        abort: Callable[[str], None] = abort_with_error,
    ):
        self.app = app
        self.database = database
        self.auto_approve = auto_approve
        self.vpc_config = vpc_config
        self.db_connection_string = db_connection_string
        self.maintenance_page_provider = maintenance_page_provider
        self.input = input
        self.echo = echo
        self.abort = abort

        if not self.app:
            if not Path(PLATFORM_CONFIG_FILE).exists():
                self.abort("You must either be in a deploy repo, or provide the --app option.")

            config = load_and_validate_platform_config()
            self.app = config["application"]

        try:
            self.application = load_application(self.app)
        except ApplicationNotFoundError:
            abort(f"No such application '{app}'.")

    def _execute_operation(self, is_dump: bool, env: str, vpc_name: str, to_env: str):
        vpc_name = self.enrich_vpc_name(env, vpc_name)

        environments = self.application.environments
        environment = environments.get(env)
        if not environment:
            self.abort(
                f"No such environment '{env}'. Available environments are: {', '.join(environments.keys())}"
            )

        env_session = environment.session

        try:
            vpc_config = self.vpc_config(env_session, self.app, env, vpc_name)
        except AWSException as ex:
            self.abort(str(ex))

        database_identifier = f"{self.app}-{env}-{self.database}"

        try:
            db_connection_string = self.db_connection_string(
                env_session, self.app, env, database_identifier
            )
        except Exception as exc:
            self.abort(f"{exc} (Database: {database_identifier})")

        try:
            task_arn = self.run_database_copy_task(
                env_session, env, vpc_config, is_dump, db_connection_string, to_env
            )
        except Exception as exc:
            self.abort(f"{exc} (Account id: {self.account_id(env)})")

        if is_dump:
            message = f"Dumping {self.database} from the {env} environment into S3"
        else:
            message = f"Loading data into {self.database} in the {env} environment from S3"

        self.echo(message, fg="white", bold=True)
        self.echo(
            f"Task {task_arn} started. Waiting for it to complete (this may take some time)...",
            fg="white",
        )
        self.tail_logs(is_dump, env)

    def enrich_vpc_name(self, env, vpc_name):
        if not vpc_name:
            if not Path(PLATFORM_CONFIG_FILE).exists():
                self.abort("You must either be in a deploy repo, or provide the vpc name option.")
            config = load_and_validate_platform_config()
            env_config = apply_environment_defaults(config)["environments"]
            vpc_name = env_config.get(env, {}).get("vpc")
        return vpc_name

    def run_database_copy_task(
        self,
        session: boto3.session.Session,
        env: str,
        vpc_config: Vpc,
        is_dump: bool,
        db_connection_string: str,
        to_env: str,
    ) -> str:
        client = session.client("ecs")
        action = "dump" if is_dump else "load"
        env_vars = [
            {"name": "DATA_COPY_OPERATION", "value": action.upper()},
            {"name": "DB_CONNECTION_STRING", "value": db_connection_string},
            {"name": "TO_ENVIRONMENT", "value": to_env},
        ]
        if not is_dump:
            env_vars.append({"name": "ECS_CLUSTER", "value": f"{self.app}-{env}"})

        response = client.run_task(
            taskDefinition=f"arn:aws:ecs:eu-west-2:{self.account_id(env)}:task-definition/{self.app}-{env}-{self.database}-{action}",
            cluster=f"{self.app}-{env}",
            capacityProviderStrategy=[
                {"capacityProvider": "FARGATE", "weight": 1, "base": 0},
            ],
            networkConfiguration={
                "awsvpcConfiguration": {
                    "subnets": vpc_config.subnets,
                    "securityGroups": vpc_config.security_groups,
                    "assignPublicIp": "DISABLED",
                }
            },
            overrides={
                "containerOverrides": [
                    {
                        "name": f"{self.app}-{env}-{self.database}-{action}",
                        "environment": env_vars,
                    }
                ]
            },
        )

        return response.get("tasks", [{}])[0].get("taskArn")

    def dump(self, env: str, vpc_name: str, to_env: str):
        self._execute_operation(True, env, vpc_name, to_env)

    def load(self, env: str, vpc_name: str):
        if self.is_confirmed_ready_to_load(env):
            self._execute_operation(False, env, vpc_name, to_env=env)

    def copy(
        self,
        from_env: str,
        to_env: str,
        from_vpc: str,
        to_vpc: str,
        services: tuple[str],
        template: str,
        no_maintenance_page: bool = False,
    ):
        to_vpc = self.enrich_vpc_name(to_env, to_vpc)
        if not no_maintenance_page:
            self.maintenance_page_provider.activate(self.app, to_env, services, template, to_vpc)
        self.dump(from_env, from_vpc, to_env)
        self.load(to_env, to_vpc)
        if not no_maintenance_page:
            self.maintenance_page_provider.deactivate(self.app, to_env)

    def is_confirmed_ready_to_load(self, env: str) -> bool:
        if self.auto_approve:
            return True

        user_input = self.input(
            f"\nWARNING: the load operation is destructive and will delete the {self.database} database in the {env} environment. Continue? (y/n)"
        )
        return user_input.lower().strip() in ["y", "yes"]

    def tail_logs(self, is_dump: bool, env: str):
        action = "dump" if is_dump else "load"
        log_group_name = f"/ecs/{self.app}-{env}-{self.database}-{action}"
        log_group_arn = f"arn:aws:logs:eu-west-2:{self.account_id(env)}:log-group:{log_group_name}"
        self.echo(f"Tailing {log_group_name} logs", fg="yellow")
        session = self.application.environments[env].session
        log_client = session.client("logs")
        wait_for_log_group_to_exist(log_client, log_group_name)
        response = log_client.start_live_tail(logGroupIdentifiers=[log_group_arn])

        stopped = False
        for data in response["responseStream"]:
            if stopped:
                break
            results = data.get("sessionUpdate", {}).get("sessionResults", [])
            for result in results:
                message = result.get("message")

                if message:
                    match = re.match(r"(Stopping|Aborting) data (load|dump).*", message)
                    if match:
                        if match.group(1) == "Aborting":
                            self.abort("Task aborted abnormally. See logs above for details.")
                        stopped = True
                    self.echo(message)

    def account_id(self, env):
        envs = self.application.environments
        if env in envs:
            return envs.get(env).account_id



PLATFORM_CONFIG_FILE = "platform-config.yml"
PLATFORM_HELPER_VERSION_FILE = ".platform-helper-version"
DEFAULT_TERRAFORM_PLATFORM_MODULES_VERSION = "5"
PLATFORM_HELPER_CACHE_FILE = ".platform-helper-config-cache.yml"

# Keys
CODEBASE_PIPELINES_KEY = "codebase_pipelines"
ENVIRONMENTS_KEY = "environments"

# Conduit
CONDUIT_ADDON_TYPES = [
    "opensearch",
    "postgres",
    "redis",
]
CONDUIT_DOCKER_IMAGE_LOCATION = "public.ecr.aws/uktrade/tunnel"



import os

from dbt_platform_helper.constants import CONDUIT_ADDON_TYPES


class ValidationException(Exception):
    pass


class PlatformException(Exception):
    pass


class AWSException(PlatformException):
    pass


class ApplicationException(PlatformException):
    pass


class CloudFormationException(AWSException):
    def __init__(self, stack_name: str, current_status: str):
        super().__init__(
            f"The CloudFormation stack '{stack_name}' is not in a good state: {current_status}"
        )


class CommitNotFoundError:
    def __init__(self, commit: str):
        super().__init__(
            f"""The commit hash "{commit}" either does not exist or you need to run `git fetch`."""
        )


class IncompatibleMajorVersion(ValidationException):
    def __init__(self, app_version: str, check_version: str):
        super().__init__()
        self.app_version = app_version
        self.check_version = check_version


class IncompatibleMinorVersion(ValidationException):
    def __init__(self, app_version: str, check_version: str):
        super().__init__()
        self.app_version = app_version
        self.check_version = check_version


class NoClusterError(AWSException):
    def __init__(self, application_name: str, environment: str):
        super().__init__(
            f"""No ECS cluster found for "{application_name}" in "{environment}" environment."""
        )


class CreateTaskTimeoutError(AWSException):
    def __init__(self, addon_name: str, application_name: str, environment: str):
        super().__init__(
            f"""Client ({addon_name}) ECS task has failed to start for "{application_name}" in "{environment}" environment."""
        )


class ParameterNotFoundError(AWSException):
    def __init__(self, application_name: str, environment: str):
        super().__init__(
            f"""No parameter called "/copilot/applications/{application_name}/environments/{environment}/addons". Try deploying the "{application_name}" "{environment}" environment."""
        )


class AddonNotFoundError(AWSException):
    def __init__(self, addon_name: str):
        super().__init__(f"""Addon "{addon_name}" does not exist.""")


class InvalidAddonTypeError(AWSException):
    def __init__(self, addon_type):
        self.addon_type = addon_type
        super().__init__(
            f"""Addon type "{self.addon_type}" is not supported, we support: {", ".join(CONDUIT_ADDON_TYPES)}."""
        )


class AddonTypeMissingFromConfigError(AWSException):
    def __init__(self, addon_name: str):
        super().__init__(
            f"""The configuration for the addon {addon_name}, is misconfigured and missing the addon type."""
        )


class CopilotCodebaseNotFoundError(PlatformException):
    def __init__(self, codebase: str):
        super().__init__(
            f"""The codebase "{codebase}" either does not exist or has not been deployed."""
        )


class NotInCodeBaseRepositoryError(PlatformException):
    def __init__(self):
        super().__init__(
            "You are in the deploy repository; make sure you are in the application codebase repository.",
        )


class NoCopilotCodebasesFoundError(PlatformException):
    def __init__(self, application_name: str):
        super().__init__(f"""No codebases found for application "{application_name}".""")


class ImageNotFoundError(PlatformException):
    def __init__(self, commit: str):
        super().__init__(
            f"""The commit hash "{commit}" has not been built into an image, try the `platform-helper codebase build` command first."""
        )


class ApplicationDeploymentNotTriggered(PlatformException):
    def __init__(self, codebase: str):
        super().__init__(f"""Your deployment for {codebase} was not triggered.""")


class ApplicationNotFoundError(ApplicationException):
    def __init__(self, application_name: str):
        super().__init__(
            f"""The account "{os.environ.get("AWS_PROFILE")}" does not contain the application "{application_name}"; ensure you have set the environment variable "AWS_PROFILE" correctly."""
        )


class ApplicationEnvironmentNotFoundError(ApplicationException):
    def __init__(self, environment: str):
        super().__init__(
            f"""The environment "{environment}" either does not exist or has not been deployed."""
        )


class SecretNotFoundError(AWSException):
    def __init__(self, secret_name: str):
        super().__init__(f"""No secret called "{secret_name}".""")


class ECSAgentNotRunning(AWSException):
    def __init__(self):
        super().__init__("""ECS exec agent never reached "RUNNING" status""")


class ResourceNotFoundException(AWSException):
    pass



import json
import stat
import subprocess
from collections.abc import Callable
from pathlib import Path

import click
import requests
import yaml
from boto3 import Session

from dbt_platform_helper.exceptions import ApplicationDeploymentNotTriggered
from dbt_platform_helper.exceptions import ApplicationEnvironmentNotFoundError
from dbt_platform_helper.exceptions import NotInCodeBaseRepositoryError
from dbt_platform_helper.utils.application import Application
from dbt_platform_helper.utils.application import load_application
from dbt_platform_helper.utils.aws import check_codebase_exists
from dbt_platform_helper.utils.aws import check_image_exists
from dbt_platform_helper.utils.aws import get_aws_session_or_abort
from dbt_platform_helper.utils.aws import get_build_url_from_arn
from dbt_platform_helper.utils.aws import list_latest_images
from dbt_platform_helper.utils.aws import start_build_extraction
from dbt_platform_helper.utils.files import mkfile
from dbt_platform_helper.utils.git import check_if_commit_exists
from dbt_platform_helper.utils.template import setup_templates


class Codebase:
    def __init__(
        self,
        input: Callable[[str], str] = click.prompt,
        echo: Callable[[str], str] = click.secho,
        confirm: Callable[[str], bool] = click.confirm,
        load_application: Callable[[str], Application] = load_application,
        get_aws_session_or_abort: Callable[[str], Session] = get_aws_session_or_abort,
        check_codebase_exists: Callable[[str], str] = check_codebase_exists,
        check_image_exists: Callable[[str], str] = check_image_exists,
        get_build_url_from_arn: Callable[[str], str] = get_build_url_from_arn,
        list_latest_images: Callable[[str], str] = list_latest_images,
        start_build_extraction: Callable[[str], str] = start_build_extraction,
        check_if_commit_exists: Callable[[str], str] = check_if_commit_exists,
        run_subprocess: Callable[[str], str] = subprocess.run,
    ):
        self.input = input
        self.echo = echo
        self.confirm = confirm
        self.load_application = load_application
        self.get_aws_session_or_abort = get_aws_session_or_abort
        self.check_codebase_exists = check_codebase_exists
        self.check_image_exists = check_image_exists
        self.get_build_url_from_arn = get_build_url_from_arn
        self.list_latest_images = list_latest_images
        self.start_build_extraction = start_build_extraction
        self.check_if_commit_exists = check_if_commit_exists
        self.run_subprocess = run_subprocess

    def prepare(self):
        """Sets up an application codebase for use within a DBT platform
        project."""
        templates = setup_templates()

        repository = (
            self.run_subprocess(
                ["git", "remote", "get-url", "origin"], capture_output=True, text=True
            )
            .stdout.split("/")[-1]
            .strip()
            .removesuffix(".git")
        )
        if repository.endswith("-deploy") or Path("./copilot").exists():
            raise NotInCodeBaseRepositoryError()

        builder_configuration_url = "https://raw.githubusercontent.com/uktrade/ci-image-builder/main/image_builder/configuration/builder_configuration.yml"
        builder_configuration_response = requests.get(builder_configuration_url)
        builder_configuration_content = yaml.safe_load(
            builder_configuration_response.content.decode("utf-8")
        )
        builder_versions = next(
            (
                item
                for item in builder_configuration_content["builders"]
                if item["name"] == "paketobuildpacks/builder-jammy-base"
            ),
            None,
        )
        builder_version = max(x["version"] for x in builder_versions["versions"])
        builder_version = min(builder_version, "0.4.240")

        Path("./.copilot/phases").mkdir(parents=True, exist_ok=True)
        image_build_run_contents = templates.get_template(f".copilot/image_build_run.sh").render()

        config_contents = templates.get_template(f".copilot/config.yml").render(
            repository=repository, builder_version=builder_version
        )
        self.echo(
            mkfile(
                Path("."), ".copilot/image_build_run.sh", image_build_run_contents, overwrite=True
            )
        )

        image_build_run_file = Path(".copilot/image_build_run.sh")
        image_build_run_file.chmod(image_build_run_file.stat().st_mode | stat.S_IEXEC)

        self.echo(mkfile(Path("."), ".copilot/config.yml", config_contents, overwrite=True))

        for phase in ["build", "install", "post_build", "pre_build"]:
            phase_contents = templates.get_template(f".copilot/phases/{phase}.sh").render()

            self.echo(
                mkfile(Path("./.copilot"), f"phases/{phase}.sh", phase_contents, overwrite=True)
            )

    def build(self, app: str, codebase: str, commit: str):
        """Trigger a CodePipeline pipeline based build."""
        session = self.get_aws_session_or_abort()
        self.load_application(app, default_session=session)

        self.check_if_commit_exists(commit)

        codebuild_client = session.client("codebuild")
        build_url = self.__start_build_with_confirmation(
            self.confirm,
            codebuild_client,
            self.get_build_url_from_arn,
            f'You are about to build "{app}" for "{codebase}" with commit "{commit}". Do you want to continue?',
            {
                "projectName": f"codebuild-{app}-{codebase}",
                "artifactsOverride": {"type": "NO_ARTIFACTS"},
                "sourceVersion": commit,
            },
        )

        if build_url:
            return self.echo(
                f"Your build has been triggered. Check your build progress in the AWS Console: {build_url}"
            )

        raise ApplicationDeploymentNotTriggered(codebase)

    def deploy(self, app, env, codebase, commit):
        """Trigger a CodePipeline pipeline based deployment."""
        session = self.get_aws_session_or_abort()

        application = self.load_application(app, default_session=session)
        if not application.environments.get(env):
            raise ApplicationEnvironmentNotFoundError(env)

        self.check_codebase_exists(session, application, codebase)

        self.check_image_exists(session, application, codebase, commit)

        codebuild_client = session.client("codebuild")
        build_url = self.__start_build_with_confirmation(
            self.confirm,
            codebuild_client,
            self.get_build_url_from_arn,
            f'You are about to deploy "{app}" for "{codebase}" with commit "{commit}" to the "{env}" environment. Do you want to continue?',
            {
                "projectName": f"pipeline-{application.name}-{codebase}-BuildProject",
                "artifactsOverride": {"type": "NO_ARTIFACTS"},
                "sourceTypeOverride": "NO_SOURCE",
                "environmentVariablesOverride": [
                    {"name": "COPILOT_ENVIRONMENT", "value": env},
                    {"name": "IMAGE_TAG", "value": f"commit-{commit}"},
                ],
            },
        )

        if build_url:
            return self.echo(
                "Your deployment has been triggered. Check your build progress in the AWS Console: "
                f"{build_url}",
            )

        raise ApplicationDeploymentNotTriggered(codebase)

    def list(self, app: str, with_images: bool):
        """List available codebases for the application."""
        session = self.get_aws_session_or_abort()
        application = self.load_application(app, session)
        ssm_client = session.client("ssm")
        ecr_client = session.client("ecr")
        codebases = self.__get_codebases(application, ssm_client)

        self.echo("The following codebases are available:")

        for codebase in codebases:
            self.echo(f"- {codebase['name']} (https://github.com/{codebase['repository']})")
            if with_images:
                self.list_latest_images(
                    ecr_client,
                    f"{application.name}/{codebase['name']}",
                    codebase["repository"],
                    self.echo,
                )

        self.echo("")

    def __get_codebases(self, application, ssm_client):
        parameters = ssm_client.get_parameters_by_path(
            Path=f"/copilot/applications/{application.name}/codebases",
            Recursive=True,
        )["Parameters"]

        codebases = [json.loads(p["Value"]) for p in parameters]

        if not codebases:
            return []
        return codebases

    def __start_build_with_confirmation(
        self,
        confirm,
        codebuild_client,
        get_build_url_from_arn,
        confirmation_message,
        build_options,
    ):
        if confirm(confirmation_message):
            build_arn = self.start_build_extraction(codebuild_client, build_options)
            return get_build_url_from_arn(build_arn)
        return None



import os
from copy import deepcopy
from datetime import datetime
from os import makedirs
from pathlib import Path

import click
import yaml
from jinja2 import Environment
from jinja2 import FileSystemLoader

from dbt_platform_helper.constants import PLATFORM_HELPER_CACHE_FILE


def to_yaml(value):
    return yaml.dump(value, sort_keys=False)


def mkfile(base_path, file_path, contents, overwrite=False):
    file_path = Path(file_path)
    file = Path(base_path).joinpath(file_path)
    file_exists = file.exists()

    if not file_path.parent.exists():
        makedirs(file_path.parent)

    if file_exists and not overwrite:
        return f"File {file_path} exists; doing nothing"

    action = "overwritten" if file_exists and overwrite else "created"

    file.write_text(contents)

    return f"File {file_path} {action}"


def generate_override_files(base_path, file_path, output_dir):
    def generate_files_for_dir(pattern):
        for file in file_path.glob(pattern):
            if file.is_file():
                contents = file.read_text()
                file_name = str(file).removeprefix(f"{file_path}/")
                click.echo(
                    mkfile(
                        base_path,
                        output_dir / file_name,
                        contents,
                        overwrite=True,
                    )
                )

    generate_files_for_dir("*")
    generate_files_for_dir("bin/*")


def generate_override_files_from_template(base_path, overrides_path, output_dir, template_data={}):
    templates = Environment(
        loader=FileSystemLoader(f"{overrides_path}"), keep_trailing_newline=True
    )
    environments = ",".join([env["name"] for env in template_data["environments"]])
    data = {"environments": environments}

    def generate_files_for_dir(pattern):

        for file in overrides_path.glob(pattern):
            if file.is_file():
                file_name = str(file).removeprefix(f"{overrides_path}/")
                contents = templates.get_template(str(file_name)).render(data)
                message = mkfile(base_path, output_dir / file_name, contents, overwrite=True)
                click.echo(message)

    generate_files_for_dir("*")
    generate_files_for_dir("bin/*")


def apply_environment_defaults(config):
    if "environments" not in config:
        return config

    enriched_config = deepcopy(config)

    environments = enriched_config["environments"]
    env_defaults = environments.get("*", {})
    without_defaults_entry = {
        name: data if data else {} for name, data in environments.items() if name != "*"
    }

    default_versions = config.get("default_versions", {})

    def combine_env_data(data):
        return {
            **env_defaults,
            **data,
            "versions": {
                **default_versions,
                **env_defaults.get("versions", {}),
                **data.get("versions", {}),
            },
        }

    defaulted_envs = {
        env_name: combine_env_data(env_data)
        for env_name, env_data in without_defaults_entry.items()
    }

    enriched_config["environments"] = defaulted_envs

    return enriched_config


def read_supported_versions_from_cache(resource_name):

    platform_helper_config = read_file_as_yaml(PLATFORM_HELPER_CACHE_FILE)

    return platform_helper_config.get(resource_name).get("versions")


def write_to_cache(resource_name, supported_versions):

    platform_helper_config = {}

    if os.path.exists(PLATFORM_HELPER_CACHE_FILE):
        platform_helper_config = read_file_as_yaml(PLATFORM_HELPER_CACHE_FILE)

    cache_dict = {
        resource_name: {
            "versions": supported_versions,
            "date-retrieved": datetime.now().strftime("%d-%m-%y %H:%M:%S"),
        }
    }

    platform_helper_config.update(cache_dict)

    with open(PLATFORM_HELPER_CACHE_FILE, "w") as file:
        file.write("# [!] This file is autogenerated via the platform-helper. Do not edit.\n")
        yaml.dump(platform_helper_config, file)


def cache_refresh_required(resource_name) -> bool:
    """
    Checks if the platform-helper should reach out to AWS to 'refresh' its
    cached values.

    An API call is needed if any of the following conditions are met:
        1. No cache file (.platform-helper-config.yml) exists.
        2. The resource name (e.g. redis, opensearch) does not exist within the cache file.
        3. The date-retrieved value of the cached data is > than a time interval. In this case 1 day.
    """

    if not os.path.exists(PLATFORM_HELPER_CACHE_FILE):
        return True

    platform_helper_config = read_file_as_yaml(PLATFORM_HELPER_CACHE_FILE)

    if platform_helper_config.get(resource_name):
        return check_if_cached_datetime_is_greater_than_interval(
            platform_helper_config[resource_name].get("date-retrieved"), 1
        )

    return True


def check_if_cached_datetime_is_greater_than_interval(date_retrieved, interval_in_days):

    current_datetime = datetime.now()
    cached_datetime = datetime.strptime(date_retrieved, "%d-%m-%y %H:%M:%S")
    delta = current_datetime - cached_datetime

    return delta.days > interval_in_days


def read_file_as_yaml(file_name):

    return yaml.safe_load(Path(file_name).read_text())
